/* cobrain, via pbrain.c */

/* pbrain interpreter in old-style C 
   This is an interpreter for Paul M. Parks's pbrain programming language,
       a variant of Urban Mueller's programming language.
   Do anything you want with this program.
   I welcome bug reports and feature requests.
   Daniel B Cristofani (http://www.hevanet.com/cristofd/)

   20130327, extended and put into service with OpenCOBOL as

   cobrain

   btiffin (https://sourceforge.net/users/btiffin)
   Tectonics: cobc -x cobrainrun.cob subcranial.cob cobrain.c -lreadline
              cobc -x cobrainmain.c subcranial.cob -lreadline
              cobc -x callcobrain.cob cobrain.c
              01 brain-cells.
                 05 brain-cell usage binary-c-long occurs 65536 times.
              CALL "cobrain" USING brain-cells
                  BY CONTENT "file pathname" & x"00"
               or
                  BY CONTENT "{cobrain source text} 99 . 10 ." & x"00"
                  RETURNING cell
              END-CALL
              DISPLAY brain-cell(cell) END-DISPLAY
*/

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <termios.h>
#include <libcob.h>

#define SIZE 65536
#define NUMBITS sizeof(long)*8
#define CA(x) case x: fprintf(stderr, "Error: " 

/* cobrain, a single character read intrepret programming tool */
void (*vectors[256])(int v);

/* colon definitions get passed source-tape offset */
int offsets[256];

/* cobrain repl has an init bloc */
int once = 1;
int bail = 0;
int retcode = 0;

long a[SIZE], memreg;
int s[SIZE], sp, t[SIZE], p, q, length, c, tmp, scale, tracer, tracing;
char code[SIZE], *f, bin[NUMBITS+1];
FILE *input;
static struct termios oldt;

const char *long_to_binary(unsigned long x) {
    unsigned long i=1UL << NUMBITS-1;
    char *d=bin;
    while (i>>=1) *d++=(i & x)?'1':'0';
    *d++ = '\0';
    return bin;
}

int e(int i){
    switch(i){
        CA(2) "call to undefined procedure (%hu) with %d at %d of %s", a[p], p, q, f); break;
        CA(3) "pointer too far %s at %d of %s", p>0?"right":"left", q, f); break;
        CA(4) "unmatched '[' at byte %d of %s", s[sp], f); break;
        CA(5) "unmatched ']' at byte %d of %s", q, f); break;
        CA(6) "unmatched '(' at byte %d of %s", s[sp], f); break;
        CA(7) "unmatched ')' at byte %d of %s", q, f); break;
        CA(8) "can't open %s", f); break;
        CA(9) "unmatched '{' at byte %d of %s", s[sp], f); break;
        CA(10) "unmatched '}' at byte %d of %s", q, f); break;
        CA(11) "unmatched '\"' at byte %d of %s", q, f); break;
        CA(12) "divide by zero at byte %d of %s", q, f); break;
    }
    printf(".\n");
    scale=0;
    sp=0;
    bail=1;
    return(i);
}


void restore_terminal_settings(void)
{
    tcsetattr(0, TCSANOW, &oldt);  /* Apply saved settings */
}

void disable_waiting_for_enter(void)
{
    struct termios newt;

    /* Make terminal read 1 char at a time */
    tcgetattr(0, &oldt);  /* Save terminal settings */
    newt = oldt;  /* Init new settings */
    newt.c_lflag &= ~(ICANON | ECHO);  /* Change settings */
    tcsetattr(0, TCSANOW, &newt);  /* Apply settings */
    atexit(restore_terminal_settings); /* Make sure settings will be restored when program ends  */
}

int opencobol(int val) {
    char module_name[256], *d=module_name;
    int (*ocmodule)(int), r;
    while ((tmp=(code[++q]!='"'))&&q<length) *d++=code[q];
    if (tmp) {e(10); return(0);}
    *d++=0;
    /* OpenCOBOL is just grand */
    //cob_init(0, NULL);  C cobrainmain.c will need the init
    ocmodule=cob_resolve(module_name);
    r=(ocmodule!=NULL)?ocmodule(val):0;
    //cob_tidy();         C cobrainmain.c will need the tidy
    return r;
} 

/* the cobrain default vectors */

/* no-op */
void implement_comment(int v) {
    //printf("comment at %d\n", q);
    scale=0;
}

/* @ fetch */
void implement_33(int v) {
    a[p]=memreg;
    scale=0;
}

/* " call module */
void implement_34(int v) {
    a[p]=opencobol(a[p]);
    scale=0;
}

/* # output formatted number */
void implement_35(int v) {
    printf("%020ld ", a[p]);
    scale = 0;
}

/* $ pointers! */
void implement_36(int v) {
    memreg=(long)&a[p];
    scale = 0;
}

/* % mod */
void implement_37(int v) {
    if (p<1) {e(3); return;}
    a[p-1] %= a[p--];
    scale = 0;
}

/* & and */
void implement_38(int v) {
    if (p<1) {e(3); return;}
    a[p-1] &= a[p--];
    scale = 0;
}

/* enter on true ( */
void implement_40(int v) {
    if (!a[p]) q=t[q];
    scale = 0;
}

/* close ) is a marker to skip to, but does nothing */
void implement_41(int v) {
    scale = 0;
}

/* * multiply */
void implement_42(int v) {
    if (p<1) {e(3); return;}
    a[p-1] *= a[p--];
    scale = 0;
}

/* + plus */
void implement_43(int v) {
    if (p<1) {e(3); return;}
    a[p-1] += a[p--];
    scale = 0;
}

/* , charater input */
void implement_44(int v) {
     if ((c = getchar()) != EOF) {
         a[p] = c == '\n' ? 10 : c;
     } else {
        printf("eof %d ", c);
        a[p]=0;
     }
     scale=0;
}

/* - minus */    
void implement_45(int v) {
    if (p<1) {e(3); return;}
    a[p-1] -= a[p--];
    scale = 0;
}

/* . character output */
void implement_46(int v) {
    putchar(a[p] == 10 ? '\n' : a[p]);
    scale = 0;
}

/* / integer divide with remainder */
void implement_47(int v) {
    if (p<1) {e(3); return;}
    if (a[p]) {
        tmp=a[p-1];
        a[p-1] /= a[p];
        a[p]=tmp - a[p] * a[p-1];
    }
    scale = 0;
}

/* digits */
void implement_digit(int v) {
    if (!scale) p++;
    a[p] *= (scale++) ? 10 : 0;
    a[p] += v;
}

/* numbers starting with zero, do not advance p, but otherwise scale in */
void implement_48(int v) {
    a[p] *= (scale++) ? 10 : 0;
}
void implement_49(int v) {
    implement_digit(1);
}
void implement_50(int v) {
    implement_digit(2);
}
void implement_51(int v) {
    implement_digit(3);
}
void implement_52(int v) {
    implement_digit(4);
}
void implement_53(int v) {
    implement_digit(5);
}
void implement_54(int v) {
    implement_digit(6);
}
void implement_55(int v) {
    implement_digit(7);
}
void implement_56(int v) {
    implement_digit(8);
}
void implement_57(int v) {
    implement_digit(9);
}

/* colon definitions */
void colon(int v) {
    s[sp++]=q;
    q = offsets[v];
    scale = 0;
}

/* colon, define next character, push q, pass execution offset to colon */
void implement_58(int v) {
    vectors[code[q+1]] = &colon;
    offsets[code[q+1]] = q+1;
    q = t[q];
    scale = 0;
}
/* semi colon, pop source tape pointer */
void implement_59(int v) {
    q=s[--sp];
    scale = 0;
}

/* Advance and retreat the memory pointer */
void implement_60(int v) {
    if (--p<0) {e(3); p=0;}
    scale = 0;
}

/* accept integer to current */
void implement_61(int v) {
    scanf("%d", &a[p]);
    scale=0;
}

/* retreat memory pointer */
void implement_62(int v) {
    if(++p>=SIZE) {e(3); p=SIZE-1;}
    scale = 0;
}
    
/* dump memory */
void implement_63(int v) {
    tmp=p;
    while (tmp-->=0)
        printf("[ %05d is %020ld ]\n", tmp+1, a[tmp+1]);
    scale = 0;
}


/* fetch to register */
void implement_64(int v) {
    memreg=a[p];
    scale = 0;
}


/* looping and conditionals */
/* skip to ] on zero */
void implement_91(int v) {
    if (!a[p]) q=t[q];
    scale = 0;
}
/* jump back to [ if not zero */
void implement_93(int v) {
    if (a[p]) q=t[q];
    scale = 0;
}

/* ^ xor */
void implement_94(int v) {
    if (p<1) {e(3); return;}
    a[p-1] ^= a[p--];
    scale = 0;
}

/* b output binary */
void implement_98(int v) {
    printf("%64s", long_to_binary(a[p]));
    scale = 0;
}

/* c compare to zero, -1 0 or 1 left in current */
void implement_99(int v) {
    a[p] = (a[p] ? ((a[p] < 0) ? -1 : 1) : 0);
    scale = 0;
}

/* d dup */
void implement_100(int v) {
    if (p>=SIZE) {e(3); return;}
    a[++p]=a[p-1];
    scale = 0;
}

/* g goback */
void implement_103(int v) {
    retcode = p+1;
    bail = 1;
    scale = 0;
}

/* l left shift */
void implement_108(int v) {
    if (p<1) {e(3); return;}
    a[p-1] <<= a[p--];
    scale = 0;
}

/* m minus */
void implement_109(int v) {
    a[p]--;
    scale = 0;
}

/* n output number */
void implement_110(int v) {
    printf("%d", a[p]);
    scale = 0;
}

/* o over */
void implement_111(int v) {
    if (p<1||p>=SIZE) {e(3); return;}
    a[++p]=a[p-2];
    scale = 0;
}

/* p plus */
void implement_112(int v) {
    a[p]++;
    scale = 0;
}

/* q quit */
void implement_113(int v) {
    retcode = -1;
    bail = 1;
    scale = 0;
}

/* r right shift */
void implement_114(int v) {
    if (p<1) {e(3); return;}
    a[p-1] >>= a[p--];
    scale = 0;
}

/* s swap */
void implement_115(int v) {
    if (p<1) {e(3); return;}
    tmp=a[p];a[p]=a[p-1];a[p-1]=tmp;
    scale = 0;
}

/* t trace */
void implement_116(int v) {
    tracer = (tracer ? 0 : 1);
    scale = 0;
}

/* { if false */
void implement_123(int v) {
    if (a[p]) q=t[q];
    scale = 0;
}

/* | or  */
void implement_124(int v) {
    if (p<1) {e(3); return;}
    a[p-1] |= a[p--];
    scale = 0;
}

/* } is skip marker and no-op */
void implement_125(int v) {
    scale = 0;
}

/* ~ not */
void implement_126(int v) {
    a[p] = ~a[p];
    scale = 0;
}

/* the cobrain interpreter */
int cobrain(char *args){

    if (once) {    
        for (tmp=0; tmp<=256; tmp++) {
            vectors[tmp] = &implement_comment;
        }

        /* ! store from register */
        vectors[33] = &implement_33;

        /* " dynamic link and call */
        vectors[34] = &implement_34;

        /* # formatted out */
        vectors[35] = &implement_35;

        /* $ fetch address of */
        vectors[36] = &implement_36;

        /* %& */
        vectors[37] = &implement_37; 
        vectors[38] = &implement_38;  
    
        /* () */
        vectors[40] = &implement_40;
        vectors[41] = &implement_41;

        /* *+ */    
        vectors[42] = &implement_42;
        vectors[43] = &implement_43;

        /* , */
        vectors[44] = &implement_44;

        /* - */
        vectors[45] = &implement_45;

        /* . */
        vectors[46] = &implement_46;

        /* / */
        vectors[47] = &implement_47;

        /* 0-9 */
        vectors[48] = &implement_48;
        vectors[49] = &implement_49;
        vectors[50] = &implement_50;
        vectors[51] = &implement_51;
        vectors[52] = &implement_52;
        vectors[53] = &implement_53;
        vectors[54] = &implement_54;
        vectors[55] = &implement_55;
        vectors[56] = &implement_56;
        vectors[57] = &implement_57;
    
        /* :; colon definitions */
        vectors[58] = &implement_58;
        vectors[59] = &implement_59;
    
        /* <> advance */
        vectors[60] = &implement_60;
    
        /* = accept integer */
        vectors[61] = &implement_61;

        /* < retreat */
        vectors[62] = &implement_62;
    
        /* ? display memory tape */
        vectors[63] = &implement_63;

        /* @ fetch */
        vectors[64] = &implement_64;

        /* [] non-zero loop blocks */
        vectors[91] = &implement_91;
        vectors[93] = &implement_93;

        /* xor */
        vectors[94] = &implement_94;

        /* b output binary */
        vectors[98] = &implement_98;

        /* c compare to 0, status in memreg */
        vectors[99] = &implement_99;

        /* d dup */
        vectors[100] = &implement_100;

        /* g goback */
        vectors[103] = &implement_103;

        /* left shift */
        vectors[108] = &implement_108;

        /* m minus */
        vectors[109] = &implement_109;

        /* number out */
        vectors[110] = &implement_110;

        /* o over */
        vectors[111] = &implement_111;

        /* p plus */
        vectors[112] = &implement_112;

        /* q quit */
        vectors[113] = &implement_113;

        /* right shift */
        vectors[114] = &implement_114;

        /* s swap */
        vectors[115] = &implement_115;
    
        /* t trace */
        vectors[116] = &implement_116;

        /* {} false blocks */
        vectors[123] = &implement_123;

        /* or */
        vectors[124] = &implement_124;

        /* close brace marker no-op */
        vectors[125] = &implement_125;

        /* not */
        vectors[126] = &implement_126;
      
        once = 0;
    }

    //disable_waiting_for_enter();

    if (strncmp("help", args, 4) == 0) {
        /* 1028, report your station */
        /* 1034, send help */
        strncpy(code, "\"cobrainhelp\"", 12);
        f="help";
        length=12;
    } else if (strncmp("file ", args, 5) == 0) {
        args += 5;
        if(!(input = fopen(f=args, "r"))) { e(8); return(0); }
        length = fread(code, 1, SIZE, input);
        fclose(input);
        memset(a, 0, SIZE);
    } else {
        f = "cobrain";
        length=strlen(args);
        strncpy(code, args, SIZE);
    }
    scale=0;
    for (q=0; q<length && !bail; q++){
        switch(code[q]){
            case '(': case '[': case '{': case ':': s[sp++]=q; break;
            case ']': if(!sp--||code[t[t[s[sp]]=q]=s[sp]]!='[') {e(5); return(0);} break;
            case ')': if(!sp--||code[t[t[s[sp]]=q]=s[sp]]!='(') {e(5); return(0);} break;
            case '}': if(!sp--||code[t[t[s[sp]]=q]=s[sp]]!='{') {e(5); return(0);} break;
            case ';': if(!sp--||code[s[sp]]!=':') {e(7); return(0);} t[s[sp]]=q; break;
        }
    }
    if (sp) e(code[s[--sp]]=='['?4:6);

    retcode = p+1;
    for (q=0; q<length && !bail; q++) {
        if (tracer && !isspace(code[q])) {
            tmp=isgraph(code[q])?code[q]:'.';
            fprintf(stderr, "[ %c at %05d was %020ld, ", (char)tmp, p, a[p]);
            tracing=1;
        }

        vectors[code[q]](code[q]); 

        if (tracing && !isspace(code[q])) {
            fprintf(stderr, " after %05d is %020ld ]\n", p, a[p]);
            tracing=0;
        }
    }

    /* COBOL is an ordinal language, first is 1 */
    return(retcode);
}
